// O[o֐
//	GlobalFunc.cpp

#include	"stdafx.h"
#include	<WinSock2.h>
#include	<iphlpapi.h>
#include	<Winnetwk.h>
#include	"GlobalFunc.h"
#include	"util.h"

#pragma comment(lib, "iphlpapi.lib")
#pragma comment(lib, "Mpr.lib")


#define ICMP_ECHO_REPLY 0
//#define ICMP_DEST_UNREACH 3
#define ICMP_ECHO_REQUEST 8
#define ICMP_TTL_EXPIRE 11

#define	ICMP_MIN	(8)
#define	ICMP_ECHO_DEFAULT_TTL	(30)
#define	ICMP_ECHO_DEFAULT_SIZE	(32)

#pragma pack(push,1)

// IP wb_
struct IPHeader {
	unsigned char	ver_hdrlen;
	unsigned char	tos;
	unsigned short	datagram_len;
	unsigned short	id;
	unsigned short	flag_flagmentoffset;
	unsigned char	ttl;
	unsigned char	proto;
	unsigned short	hdr_chksum;
	unsigned long	source_addr;
	unsigned long	dest_addr;
};

// ICMP wb_
struct ICMPHeader {
	unsigned char	type;
	unsigned char	code;
	unsigned short	chksum;
	unsigned short	id;
	unsigned short	seq;
};

#pragma pack(pop)

unsigned char	ICMP_SendBuf[1024 + sizeof(ICMPHeader)];
unsigned char	ICMP_RecvBuf[1024 + sizeof(IPHeader)];
static int		ICMP_PacketSize;


// `FbNTvZ
unsigned short CalcChecksum(void* pData, int nLen)
{
	unsigned long	chksum = 0;
	unsigned short	*sp = (unsigned short *)pData;
	int				i = nLen;

	while (i >= 2) {
		chksum += *sp++;
		i -= sizeof(unsigned short);
	}
	if (i > 0) {
		chksum += *((unsigned char*)sp);
	}
	chksum = (chksum >> 16) + (chksum & 0xffff);
	chksum += (chksum >> 16);

	return (unsigned short)(~chksum);
}

//ICMPM̏
SOCKET	ICMP_Init(int nTtl, int nPacketSize, DWORD *pLastError)
{
	SOCKET			sock;
	ICMPHeader		*pIcmpHdr;
	unsigned char	*send_buf = &ICMP_SendBuf[0];
	unsigned long	block_mode = 1;

	if (nTtl < 0) {
		nTtl = ICMP_ECHO_DEFAULT_TTL;
	}
	if (nPacketSize < 0) {
		nPacketSize = ICMP_ECHO_DEFAULT_SIZE;
	}
	if (nPacketSize > 1024) {
		if (pLastError) {
			*pLastError = (DWORD)-1;
		}
		return INVALID_SOCKET;
	}

	// ICMPwb_Zbg
	pIcmpHdr = (ICMPHeader *)send_buf;

	pIcmpHdr->type = ICMP_ECHO_REQUEST;
	pIcmpHdr->code = 0;
	pIcmpHdr->chksum = 0;
	pIcmpHdr->id = (USHORT)GetCurrentProcessId();
	pIcmpHdr->seq = 0;

	// f[^Ƃ0xcc (1010 1010) Zbg
    memset(&send_buf[sizeof(ICMPHeader)],
					0xcc, nPacketSize - sizeof(ICMPHeader));

	// `FbNTvZ
	pIcmpHdr->chksum = CalcChecksum(pIcmpHdr, nPacketSize);

	//\Pbg
	sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (sock == INVALID_SOCKET) {
		if (pLastError) {
			*pLastError = GetLastError();
		}
		return INVALID_SOCKET;
	}
	// TTLݒ
	if (setsockopt(sock, IPPROTO_IP, IP_TTL,
						(char*)&nTtl, sizeof(int)) == SOCKET_ERROR) {
		if (pLastError) {
			*pLastError = GetLastError();
		}
		closesocket(sock);
		return INVALID_SOCKET;
	}
	//mubLO[h
	if (ioctlsocket((int)sock, FIONBIO, &block_mode) != 0) {
		if (pLastError) {
			*pLastError = GetLastError();
		}
		closesocket(sock);
		return INVALID_SOCKET;
	}
	ICMP_PacketSize = nPacketSize;

	return sock;
}

//ICMPM
int	ICMP_Check(SOCKET sock, LPCTSTR pHost, int nTimeout)
{
	sockaddr_in 	dst_addr, src_addr;
	unsigned long	ip_addr;
	hostent			*hp;
	IPHeader		*pIpHdr;
	ICMPHeader		*pIcmpHdr;
	unsigned char	*send_buf = &ICMP_SendBuf[0];
	unsigned char	*recv_buf = &ICMP_RecvBuf[0];
	int				packet_size = ICMP_PacketSize;
	int				len, hdr_len;
	int				recv_len, from_len;

	DWORD dwStartTick = ::GetTickCount();

	//̃AhXZbg
    memset(&dst_addr, 0, sizeof(dst_addr));
    dst_addr.sin_family = AF_INET;
	ip_addr = inet_addr(pHost);
	if (ip_addr == INADDR_NONE) {
		hp = gethostbyname(pHost);
		if (!hp) {
			//AhXȂ
			TRACE("ICMP_Check(): gethostbyname() error: host=%s\n", pHost);
			return -1;
		}
		memcpy(&dst_addr.sin_addr, hp->h_addr, hp->h_length);
	} else {
		dst_addr.sin_addr.s_addr = ip_addr;
	}
	// ICMPpPbgM
	len = sendto(sock, (char*)send_buf, packet_size, 0, 
									(sockaddr*)&dst_addr, sizeof(dst_addr));
	if (len == SOCKET_ERROR) {
		TRACE("ICMP_Check(): sendto() error: host=%s\n", pHost);
		return -1;
    }
	// ICMPM
	fd_set			fds;
	struct timeval	tv;
	int				res;

	tv.tv_sec = nTimeout / 1000;
	tv.tv_usec = (nTimeout % 1000) * 1000;

	int ret = -1;
	while (true) {
		//M҂
		FD_ZERO(&fds);
		FD_SET(sock, &fds);
		if ((res = select(sock + 1, &fds, NULL, NULL, &tv)) > 0) {
			if (FD_ISSET(sock, &fds)) {
				//f[^M
				from_len = sizeof(src_addr);
				len = packet_size + sizeof(IPHeader);
				recv_len = recvfrom(sock, (char *)recv_buf,
										len, 0, (sockaddr*)&src_addr, &from_len);
				if (recv_len == SOCKET_ERROR) {
					//MG[
					int	err = WSAGetLastError();
					if (err != WSAEWOULDBLOCK) {
						TRACE("ICMP_Check(): recvfrom() ERROR: host=%s\n", pHost);
						//break;
						continue;
					} 
				} else {
					//AhX`FbNi肩̉j
					if (dst_addr.sin_addr.S_un.S_addr != src_addr.sin_addr.S_un.S_addr) {
						//̃zXg̃pPbg
						TRACE("ICMP_Check(): src address %s != %s\n", inet_ntoa(src_addr.sin_addr), pHost);
						continue;
					}
					//肩̉
					pIpHdr = (IPHeader *)recv_buf;
					hdr_len = (pIpHdr->ver_hdrlen & 0x0f) * 4;
					pIcmpHdr = (ICMPHeader *)&recv_buf[hdr_len];
					if (pIcmpHdr->seq == 0) {
						if (recv_len < hdr_len + ICMP_MIN) {
							//pPbgTCYZ
							TRACE("ICMP_Check(): packet size ERROR: host=%s\n", pHost);
							ret = -1;
							break;
						} else {
							//pPbgTCY͐
							if (pIcmpHdr->type != ICMP_ECHO_REPLY) {
								//ICMPł͂Ȃ
								if (pIcmpHdr->type != ICMP_TTL_EXPIRE) {
									TRACE("ICMP_Check(): icmp type != ICMP_ECHO_REPLY: host=%s\n", pHost);
									break;
								}
								TRACE("ICMP_Check(): icmp type ICMP_TTL_EXPIRE: host=%s\n", pHost);
							} else {
								//ICMP
								if (pIcmpHdr->id == (unsigned short)GetCurrentProcessId()) {
									//ȉM
									ret = 0;
									break;
								} else {
									TRACE("ICMP_Check(): icmp id != ProcessID: host=%s\n", pHost);
									continue;
								}
							}
						}
					} else {
						//V[PXԍႤ
						TRACE("ICMP_Check(): packet seq != 0: host=%s\n", pHost);
						continue;
					}
				}
			} else {
				TRACE("ICMP_Check(): select set other socket: host=%s\n", pHost);
				continue;
			}
		} else if (res == 0) {
			//M^CAEg
			DWORD dwEndTick = ::GetTickCount();
			TRACE("ICMP_Check(): select timeout: host=%s time=%lu\n", pHost, dwEndTick - dwStartTick);
			break;
		} else {
			TRACE("ICMP_Check(): select error: host=%s\n", pHost);
			break;
		}
	}
	return ret;
}

//ICMPM̏I
void ICMP_End(SOCKET sock)
{
	closesocket(sock);
}

//ICMPpPbg𑗂ĉ`FbNi1Ȁ1֐ōsj
int	ICMP_Echo(LPCTSTR pHost, int nTtl, int nSize, int nTimeout)
{
	SOCKET 	sock;
	int		ret;

	if ((sock = ICMP_Init(nTtl, nSize, NULL)) == INVALID_SOCKET) {
		return -1;
	}
	ret = ICMP_Check(sock, pHost, nTimeout);

	ICMP_End(sock);

	return ret;
}

//zXgAhXW
int	CollectHostAddr(LPCTSTR sIpAddr, LPCTSTR sNetMask, int nTimeout,
								BOOL bNoExternCmd, CStringList& slHostList)
{
	int				i;
	SOCKET			sock = INVALID_SOCKET;
	DWORD			dwIpAddr, dwNetMask, dwAddrCount;
	DWORD			dwErr;
	char			szIpAddr[32];
	int				ret = 0;

	if (!sIpAddr || !(*sIpAddr) || !sNetMask || !(*sNetMask)) {
		return -1;
	}
	// IPAhXoCiɕϊ
	dwIpAddr = StrToDwIpAddr(sIpAddr);

	// lbg}XNoCiɕϊ
	dwNetMask = StrToDwIpAddr(sNetMask);

	//lbg}XN𔽓]ălbg[ÑAhX𓾂
	dwAddrCount = ~dwNetMask;

	// IPAhXƃlbg}XNlbg[NEAhX𐶐
	dwIpAddr = dwIpAddr & dwNetMask;

	slHostList.RemoveAll();

	if (bNoExternCmd) {
		if ((sock = ICMP_Init(-1, -1, &dwErr)) == INVALID_SOCKET) {
			return -1;
		}
	}
	for (i = 0; i < (int)dwAddrCount; i++) {
		dwIpAddr++;
		DwToStrIpAddr(dwIpAddr, szIpAddr, sizeof(szIpAddr));
	
		if (sock != INVALID_SOCKET) {
			//Õ\PbgICMP𑗂iAdministrators[U[݂̂gpłj
			ret = ICMP_Check(sock, szIpAddr, nTimeout);
		} else {
			//ping R}hĂяoiAdministratorsłȂ[U[gpꍇj
			ret = PingCheck(szIpAddr, nTimeout);
		}
		if (ret == 0) {
			slHostList.AddTail(szIpAddr);
		}
	}
	if (sock != INVALID_SOCKET) {
		ICMP_End(sock);
	}
	return slHostList.GetCount();
}


// pingR}hgċN`FbNs
int	PingCheck(LPCTSTR lpszHost, int Timeout)
{
	TCHAR	szCmd[2048];

	sprintf_s(szCmd, sizeof(szCmd), "ping -n 1 -w %d %s", Timeout, lpszHost);

	// OR}hs
	if (RunCommand(szCmd, FALSE, TRUE) != 0) {
		return -1;
	}
	return 0;
}

// }WbNEpPbgM
int	SendWakeupPacket(LPCTSTR sIpAddr, LPCTSTR sNetMask, LPCTSTR sMacAddr,
										BOOL bBloadcast, u_short nPort /*=7*/)
{
	DWORD				dwIpAddr, dwNetMask;
	int					sock;
	struct sockaddr_in	addr;
	int					addr_len;
	unsigned char		maddr[6];
	unsigned char		buff[6 * 17], *p;
	BOOL				optval;
	int					i, len, send_len;
	int					res = 0;

	ASSERT(sIpAddr && (*sIpAddr));

	if (!sIpAddr || !(*sIpAddr)) {
		// IPAhXZbgĂȂ
		return -1;
	}
	if (StrMacAddrToBin((char *)sMacAddr, maddr) != 0) {
		// sMACAhX
		return -2;
	}
	// IPAhXoCiɕϊ
	dwIpAddr = StrToDwIpAddr(sIpAddr);

	/* create socket */
	if ((sock = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		return -3;
	}
	if (bBloadcast) {
		optval = TRUE;
		if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST, (char *)&optval, sizeof(BOOL)) != 0) {
			res = -4;
		}
		// lbg}XNoCiɕϊ
		dwNetMask = StrToDwIpAddr(sNetMask);

		// IPAhXƃlbg}XNu[hLXgEAhX𐶐
		dwIpAddr = dwIpAddr | ~dwNetMask;
	}
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(nPort);
	addr.sin_addr.S_un.S_addr = htonl(dwIpAddr);

	if (res == 0) {
		//}WbNpPbg̃f[^쐬
		memset(buff, 0xff, 6);
		p = &buff[6];
		for (i = 0; i < 16; i++) {
			memcpy(p, maddr, 6);
			p += 6;
		}
		len = 6 * 17;

		addr_len = sizeof(addr);
		send_len = sendto(sock, (char *)buff, len, 0,
							(struct sockaddr *)&addr, addr_len);

		if (send_len < 0) {
			res = -5;
		}
	}
	closesocket(sock);

	return res;
}

// DWORDIPAhX𕶎ɕϊ
LPCTSTR DwToStrIpAddr(DWORD dwIpAddr, LPTSTR sAddr, int MaxLen)
{
	char		*pAddr;

	*sAddr = '\0';

	dwIpAddr = htonl(dwIpAddr);
	pAddr = inet_ntoa(*((struct in_addr *)&dwIpAddr));
	if (pAddr) {
		strncpy_s(sAddr, MaxLen, pAddr, MaxLen);
	}
	return sAddr;
}

// IPAhXDWORDɕϊ
DWORD	StrToDwIpAddr(LPCTSTR strIpAddr)
{
	DWORD	dwAddr;

	dwAddr = (DWORD)ntohl(inet_addr(strIpAddr));

	return dwAddr;
}

// MACAhXoCiɕϊ
int		StrMacAddrToBin(LPCTSTR pMacAddrStr, unsigned char *pBinMacAddr)
{
	LPCTSTR			p = pMacAddrStr;
	unsigned char	*q = pBinMacAddr;
	int				i, count;
	unsigned char	c;

	count = 0;
	i = 0;
	while (*p) {
		if (isxdigit(*p)) {
			if (i >= 2) {
				i = 0;
			}
			if ('0' <= *p && *p <= '9') {
				c = (unsigned char)(*p - '0');
			} else if ('a' <= *p && *p <= 'f') {
				c = (unsigned char)(*p - 'a' + 0x0a);
			} else if ('A' <= *p && *p <= 'F') {
				c = (unsigned char)(*p - 'A' + 0x0a);
			} else {
				return -1;
			}
			if (i == 0) {
				*q = (c << 4) & 0xf0;
			} else {
				*q |= c & 0x0f;
				q++;
				count++;
			}
			i++;
		} else {
			i = 0;
		}
		p++;
	}
	if (count != 6) {
		return -1;
	}
	return 0;
}

// oCiMACAhX𕶎ɕϊ
LPCTSTR BinMacAddrToStr(unsigned char *pBinMacAddr, LPTSTR pMacAddrStr, int MaxLen)
{
	sprintf_s(pMacAddrStr, MaxLen, "%02X:%02X:%02X:%02X:%02X:%02X",
			pBinMacAddr[0], pBinMacAddr[1], pBinMacAddr[2],
			pBinMacAddr[3], pBinMacAddr[4], pBinMacAddr[5]);

	return pMacAddrStr;
}

// zXgIPAhX擾
DWORD	GetDwMyIpAddr()
{
	struct hostent	*hp;
	char			hostname[256];
	DWORD			dwMyAddr = 0;

	if (gethostname(hostname, sizeof(hostname)) == 0) {
		if ((hp = gethostbyname(hostname))) {
			dwMyAddr = ntohl(*((unsigned long *)hp->h_addr));
			return dwMyAddr;
		}
	}
	return (DWORD)0;
}

//lbg[NA_v^̃AhX̃Xg擾
int	GetAdapterAddrList(struct sockaddr_in *addr, int max_num)
{
	ULONG	len = 0;
	DWORD	dwRes;
	int		idx = 0;

	//lbg[NA_v^̃AhXꗗ̃TCY`FbN
	if ((dwRes = GetAdaptersAddresses(AF_UNSPEC, 0, NULL, NULL, &len)) == ERROR_BUFFER_OVERFLOW) {
		//lbg[NA_v^̃AhXꗗ̃obt@mۂ
		IP_ADAPTER_ADDRESSES	*pAdptList = (IP_ADAPTER_ADDRESSES *)new BYTE[len];

		//lbg[NA_v^̃AhXꗗ擾
		if ((dwRes = GetAdaptersAddresses(AF_UNSPEC, 0, NULL, pAdptList, &len)) == ERROR_SUCCESS) {
			IP_ADAPTER_ADDRESSES	*pAdpt;

			//A_v^Ƀ`FbN
			for (pAdpt = pAdptList; pAdpt != NULL; pAdpt = pAdpt->Next) {
				//AhXȂ̂ƃ[vobNȊO
				if (pAdpt->OperStatus == IfOperStatusUp && pAdpt->PhysicalAddressLength > 0 &&
														pAdpt->IfType != IF_TYPE_SOFTWARE_LOOPBACK) {
					//jLXg̃AhXׂă`FbN
					IP_ADAPTER_UNICAST_ADDRESS	*pUniAddr;
					for (pUniAddr = pAdpt->FirstUnicastAddress; pUniAddr != NULL; pUniAddr = pUniAddr->Next) {
						//AvP[VgׂłȂAhXȊO
						if (!(pUniAddr->Flags & IP_ADAPTER_ADDRESS_TRANSIENT)) {
							//AhXIPAhX
							if (pUniAddr->Address.lpSockaddr->sa_family == AF_INET) {
								if (addr) {
									if (idx < max_num) {
										//AhX擾
										memcpy(&addr[idx], (struct sockaddr_in *)pUniAddr->Address.lpSockaddr, sizeof(struct sockaddr_in));
										idx++;
									} else {
										break;
									}
								} else {
									//AhX擾p̃obt@NULL̏ꍇ͐JEg
									idx++;
								}
							}
						}
					}
					if (addr != NULL && idx >= max_num) {
						break;
					}
				}
			}
		}
		delete[] pAdptList;
	}
	return idx;	//AhXԂ
}

//IPAhXɑΉlbg[NA_v^MACAhX擾
LPCTSTR GetAdapterMacAddr(LPCTSTR sIpAddr, LPTSTR sMacAddr, int MaxLen)
{
	//IPAhXꍇARPe[uɂ͖̂Ńlbg[NA_v^[̃AhX𒲂ׂ
	ULONG	len = 0;
	DWORD	dwRes;
	BOOL	ret = FALSE;

	//lbg[NA_v^̃AhXꗗ̃TCY`FbN
	if ((dwRes = GetAdaptersAddresses(AF_UNSPEC, 0, NULL, NULL, &len)) == ERROR_BUFFER_OVERFLOW) {
		//lbg[NA_v^̃AhXꗗ̃obt@mۂ
		IP_ADAPTER_ADDRESSES	*pAdptList = (IP_ADAPTER_ADDRESSES *)new BYTE[len];

		//lbg[NA_v^̃AhXꗗ擾
		if ((dwRes = GetAdaptersAddresses(AF_UNSPEC, 0, NULL, pAdptList, &len)) == ERROR_SUCCESS) {
			IP_ADAPTER_ADDRESSES	*pAdpt;

			//A_v^Ƀ`FbN
			for (pAdpt = pAdptList; pAdpt != NULL; pAdpt = pAdpt->Next) {
				//AhXȂ̂ƃ[vobNȊO
				if (pAdpt->OperStatus == IfOperStatusUp && pAdpt->PhysicalAddressLength > 0 &&
														pAdpt->IfType != IF_TYPE_SOFTWARE_LOOPBACK) {
					//jLXg̃AhXׂă`FbN
					IP_ADAPTER_UNICAST_ADDRESS	*pUniAddr;
					for (pUniAddr = pAdpt->FirstUnicastAddress; pUniAddr != NULL; pUniAddr = pUniAddr->Next) {
						//AvP[VgׂłȂAhXȊO
						if (!(pUniAddr->Flags & IP_ADAPTER_ADDRESS_TRANSIENT)) {
							//AhXIPAhX
							if (pUniAddr->Address.lpSockaddr->sa_family == AF_INET) {
								//AhX𕶎ɕϊ
								struct sockaddr_in *pSocAddr = (struct sockaddr_in *)pUniAddr->Address.lpSockaddr;
								char	*pAddr = inet_ntoa(*((struct in_addr *)&pSocAddr->sin_addr));
								if (strcmp(pAddr, sIpAddr) == 0) {
									//IPAhXv̂ŁAA_v^̕AhXԂ
									BinMacAddrToStr(pAdpt->PhysicalAddress, sMacAddr, MaxLen);
									ret = TRUE;
									break;
								}
							}
						}
					}
					if(ret) {
						//̂ő̃A_v^̓`FbNȂ
						break;
					}
				}
			}
		}
		delete[] pAdptList;
		if (ret) {
			return sMacAddr;
		}
	}
	return NULL;
}

//IPAhXMACAhX擾
LPCTSTR IpAddrToMacAddr(LPCTSTR sIpAddr, LPTSTR sMacAddr, int MaxLen)
{
	DWORD			dwIpAddr;
	IPAddr			ipAddr;
	char			macaddr_str[32];
	unsigned char	macaddr_bin[6];
	ULONG			ulMacAddrLen = 6;
	DWORD			dwRes;
	char			ipaddr_str[32];

	// IPAhXǨ`ɐ
	dwIpAddr = StrToDwIpAddr(sIpAddr);
	
	DwToStrIpAddr(dwIpAddr, ipaddr_str, sizeof(ipaddr_str));

	//IPAhX̏ꍇARPe[uɂ͖̂Ńlbg[NA_v^[̃AhX𒲂ׂ
	if (GetAdapterMacAddr(ipaddr_str, macaddr_str, sizeof(macaddr_str))) {
		//̃AhX̂ICMP̕Kv͖
		strcpy_s(sMacAddr, MaxLen, macaddr_str);
		return (LPCTSTR)sMacAddr;
	}

	ipAddr = inet_addr(ipaddr_str);

	memset(macaddr_bin, 0, sizeof(macaddr_bin));

	dwRes = SendARP(ipAddr, 0, (ULONG *)macaddr_bin, &ulMacAddrLen);
	if (dwRes != NO_ERROR) {
		//ARPR}h̏o͂MACAhX܂
		return NULL;
	}
	// ɕϊi؂蕶':'ɂȂ悤ɃtH[}bg𐮂j
	BinMacAddrToStr(macaddr_bin, macaddr_str, (int)sizeof(macaddr_str));

	strcpy_s(sMacAddr, MaxLen, macaddr_str);

	return (LPCTSTR)sMacAddr;
}

//IPAhXlbg}XN擾
DWORD	GetDefaultNetMask(DWORD IpAddr)
{
	if (!(IpAddr & 0x80000000)) {
		return 0xff000000;
	}
	if ((IpAddr & 0xc0000000) == 0x80000000) {
		return 0xffff0000;
	}
	if ((IpAddr & 0xe0000000) == 0xc0000000) {
		return 0xffffff00;
	}
	return 0xffffff00;
}

//IPAhXzXg𓾂
LPTSTR	GetHostNameFromIpAddr(LPCTSTR szIpAddr, LPTSTR pName, int MaxLen)
{
	struct hostent	*hp;
	struct in_addr addr;

	*pName = '\0';

	addr.S_un.S_addr = inet_addr(szIpAddr);

	hp = gethostbyaddr((const char *)&addr, sizeof(addr), AF_INET);
	if (hp) {
		strncpy_s(pName, MaxLen, hp->h_name, MaxLen);
	}
	return pName;
}
